import os

symdir = "/".join(__file__.split("/")[:-1])
currentdir = os.getcwd().split('/')[-1]

def inject_sympy_dockerfile(tag):
    injection = [
        f"FROM {tag}",
        f"ADD sympy /root/sympy",
        f"RUN pip3 install /root/sympy",
        f"WORKDIR /node/",
        f"ENV PYTHONUNBUFFERED 1",
    ]
    injectstr = '\n'.join(injection)
    os.system(f"docker build -q -t {tag} {symdir} -f-<<EOF\n{injectstr}\nEOF")

def build(dockerfile):
    tag = f"sym/{currentdir}.{dockerfile}"
    print(f"Building {tag}")
    os.system(f"docker build -t {tag} -f env/{dockerfile} {os.getcwd()}")
    inject_sympy_dockerfile(tag)
    return tag


def build_on_platform(dockerfile, platforms):
    tag = f"sym/{currentdir}.{dockerfile}"
    platforms = ",".join(platforms)
    print(f"Building {tag}")
    os.system(f"docker buildx build --platform {platforms} -t {tag} -f env/{dockerfile} {os.getcwd()}")
    inject_sympy_dockerfile(tag)
    return tag

def build_node_image(image, node, tag, platforms):
    injection = [
        f"FROM {image}",
        f"COPY src/{node} /node",
        f"COPY protos /node/protos",
    ]
    injectstr = '\n'.join(injection)
    name = f"{image}.{node}"
    print(f"Building {name}:{tag}")
    platforms = ",".join(platforms)
    os.system(f"docker buildx build --platform {platforms} -q -t {name}:{tag} . -f-<<EOF\n{injectstr}\nEOF")



import os
import sys
from glob import glob
import importlib


def get_protos():
    sys.path.append("/protos")
    sources = dict()
    for source in os.listdir("/protos"):
        sources[source] = {
            "protos": dict(),
            "msgs": dict()
        }

        for proto in glob(f"/protos/{source}/proto/*pb2.py"):
            proto_name = proto.split("/")[-1][:-3]
            module = ".".join(proto.split("/")[-3:])[:-3]
            sources[source]["protos"][proto_name] = importlib.import_module(module + "_grpc")
            sources[source]["msgs"][proto_name] = importlib.import_module(module)

    for source_name, source in sources.items():
        source["stubs"] = dict()
        source["servicers"] = dict()
        for proto_name, proto in source["protos"].items():
            for var in dir(proto):
                if var.endswith("Stub"):
                    methods = dict()
                    for m in getattr(source["msgs"][proto_name], "_" + var[:-4].upper()).methods:
                        methods[m.name] = {
                            "input": getattr(source["msgs"][proto_name], m.input_type.name),
                            "output": getattr(source["msgs"][proto_name], m.output_type.name)
                        }

                    source["stubs"][var[:-4]] = {
                        "stub": getattr(proto, var),
                        "methods": methods
                    }
                elif var.endswith("Servicer"):
                    methods = dict()
                    for m in getattr(source["msgs"][proto_name], "_" + var[:-8].upper()).methods:
                        methods[m.name] = {
                            "input": getattr(source["msgs"][proto_name], m.input_type.name),
                            "output": getattr(source["msgs"][proto_name], m.output_type.name)
                        }

                    source["servicers"][var[:-8]] = {
                        "servicer": getattr(proto, var),
                        "add_to_server": getattr(proto, f"add_{var}_to_server"),
                        "methods": methods
                    }
    return sources


import os
from symphonai import docker
import shutil
import json
import yaml
from glob import glob
import subprocess

symdir = "/".join(__file__.split("/")[:-1])
currentdir = os.getcwd().split('/')[-1]

class CLI:
    """
    A tool to build and run dockerized nodes connected with gRPC in a clear project structure.
    """
    def setup(self):
        """
        Setup project structure
        """
        os.makedirs("env", exist_ok=True)
        os.makedirs("src", exist_ok=True)
        os.makedirs("protos", exist_ok=True)
        os.makedirs(".sym", exist_ok=True)

        if not os.path.exists("global.env"):
            open("global.env", "w").close()

    def create(self, name, env):
        """
        Create a new package with name <name> and dockerfile <env>
        """
        print("Creating package", name, "with env", env)

        pkg_dir = f"src/{name}"
        if os.path.exists(pkg_dir):
            print(f"Package of name {name} already exists")
            return

        if not os.path.exists(f"env/{env}"):
            print(f"Env {env} does not exist. Creating from default env")
            shutil.copy(f"{symdir}/envs/default", f"env/{env}")

        shutil.copytree(f"{symdir}/template", f"{pkg_dir}")

        with open(f"{pkg_dir}/config.json", "w") as f:
            json.dump({
                "env": env,
                "options": {},
            }, f, indent=4)

    def build(self):
        """
        Build all docker images and create docker-compose.yml
        """
        for dockerfile in os.listdir("env"):
            docker.build(dockerfile)

        pkgs = {}
        for pkg in os.listdir("src"):
            pkgs[pkg] = json.load(open(f"src/{pkg}/config.json"))

        services = dict()
        for pkg, config in pkgs.items():
            services[pkg] = config["options"]
            services[pkg].update({
                "image": f"sym/{currentdir}.{config['env']}",
                "volumes": [f"{os.getcwd()}/protos:/node/protos", f"{os.getcwd()}/src/{pkg}:/node"],
                "networks": [currentdir],
                "env_file": [f"{os.getcwd()}/global.env", f"{os.getcwd()}/src/{pkg}/params.env"],
                "environment": [f"SYMNAME={pkg}", "PYTHONPYCACHEPREFIX=../pycache"],
                "extra_hosts": [f"symhost:host-gateway"],
            })

        with open(".sym/docker-compose.yml", "w") as f:
            yaml.dump({
                "version": "3.7",
                "services": services,
                "networks": {
                    currentdir: {
                        "driver": "bridge",
                    },
                }
            }, f)

    def run(self, profile=None):
        """
        Run the docker containers
        """
        if profile:
            os.system(f"docker compose -f .sym/docker-compose.yml --profile {profile} -p sym/{currentdir} up --remove-orphans")
        else:
            os.system(f"docker compose -f .sym/docker-compose.yml -p sym-{currentdir} up --remove-orphans")

    def stop(self):
        """
        Stop the docker containers
        """
        os.system(f"docker compose -f .sym/docker-compose.yml -p sym-{currentdir} down")
        
        
    def logs(self, *containers, follow=False, tail=100):
        """
        Stop the docker containers
        """
        
        containers = " ".join(containers)
        cmd = f"docker compose -f .sym/docker-compose.yml -p sym-{currentdir} logs {containers} --tail {tail}"
        if follow:
            cmd += " --follow"
        print(cmd)
        os.system(cmd)

    def docs(self):
        """
        Generate documentation for all protos
        """
        for proto in glob(f"protos/*.proto"):
            package = proto.split("/")[-1].split(".")[0]
            os.system(f'docker run --rm -v {os.getcwd()}/docs:/out -v {os.getcwd()}/protos:/protos/protos pseudomuto/protoc-gen-doc */{package}.proto -I / --doc_opt=markdown,{package}.md')

    def commit(self, registry, tag="latest"):
        """
        Build the docker images with code for linux/arm64 and linux/amd64
        """
        platforms = ["linux/arm64", "linux/amd64"]
        # check if docker buildx platforms are installed
        if "linux/arm64" not in os.popen("docker buildx inspect").read():
            print("Please install docker buildx for linux/arm64. ")
            print("Run <docker run --privileged --rm tonistiigi/binfmt --install all> to install emulators on your system.")
            print("Then, run <docker buildx create --use --name sym> to create a builder instance.")
            print("Finally, run <docker buildx inspect --bootstrap> to install the platforms on your system.")
            return
        
        if "linux/amd64" not in os.popen("docker buildx inspect").read():
            print("platform linux/amd64 is not available on your system. Only building for linux/arm64")
            platforms = ["linux/arm64"]
        
        envs = {}
        for dockerfile in os.listdir("env"):
            envtag = docker.build_on_platform(dockerfile, platforms)
            envs[dockerfile] = envtag
        
        pkgs = {}
        for pkg in os.listdir("src"):
            pkgs[pkg] = json.load(open(f"src/{pkg}/config.json"))
            
        for pkg, config in pkgs.items():
            image = envs[config["env"]]
            docker.build_node_image(image, pkg, tag, platforms)
        
        
        



import os
import traceback
import json
import grpc

def get_param(varname):
    try:
        return json.loads(os.getenv(varname))
    except:
        return os.getenv(varname)

import grpc
import grpc.experimental
def get_proto(package):
    return grpc.protos_and_services(f"protos/{package}.proto")

from concurrent import futures
def server(**kwargs):
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10), **kwargs)
    server.add_insecure_port('[::]:80')
    return server

def handle_exceptions_here(func):
    def handler(self, request, context):
        try:
            return func(self, request, context)
        except Exception as e:
            traceback.print_exc()
            
            context.abort(grpc.StatusCode.CANCELLED, f"There was an exception in {type(self).__name__}: {type(e).__name__}")
    return handler



import setuptools

setuptools.setup(
    name='symphonaipy',  
    version='0.2',
    author="Sven Schultze",
    author_email="schultze.sven@googlemail.com",
    description="A tool to communicate within symphonai",
    long_description="long_description",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[

    ]
)



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name='symphonai',
    version='0.2.4',
    scripts=['sym'],
    author="Sven Schultze",
    author_email="schultze.sven@googlemail.com",
    description="A tool to build and run docker images together",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_namespace_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe = False,
    #package_data={
    #    'symphonai/envs': ['*'],
    #    'symphonai/mosquitto': ['*'],
    #    'symphonai/nginx': ['*'],
    #    'symphonai/sympy': ['*'],
    #    'symphonai/template': ['*'],
    #},
    
    package_data={
    	"symphonai": ["*"],
    	"symphonai.mosquitto": ["*", "config/*"],
    	"symphonai.nginx": ["*"],
    	"symphonai.envs": ["*"],
    	"symphonai.sympy": ["*"],
    	"symphonai.template": ["*"]
    },
    install_requires=[
        'pyyaml',
        'fire'
    ]
)

